/*********************************************************************************************
Sample program for OpenGL vertex command simulation
Promit Roy
Created: 9/27/02
Last Modified: 9/28/02

Source file for the Vertex Command Emulator. Just compile this file with your project and
include the header d3dverts.h to use these functions.
*********************************************************************************************/
#include "d3dverts.h"

//Implementation file for the Direct3D vertex commands

//Definition for our vertex structure
//You can move this to the header if you want to
struct ListVertex
{
	float x, y, z;
	float nx, ny, nz;
	D3DCOLOR Diffuse;
	float tu, tv;
};
#define FVF_LISTVERTEX		D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX0

//private static variables
//The header doesn't extern these becuase, frankly, you shouldn't mess
//with them outside this file
static BOOL bRenderBegun =  FALSE;

static IDirect3DDevice8* pD3DDevice = NULL;
static D3DPRIMITIVETYPE PrimitiveType = D3DPT_TRIANGLELIST;

static ListVertex* pVertList = NULL;
static long NumVerts = 0;
static long MaxArraySize = 0;

static float CurNX = 0.0, CurNY = 0.0, CurNZ = 0.0;
static D3DCOLOR CurDiffuse = 0xff000000;		// = D3DCOLOR_XRGB( 0, 0, 0 )
static float CurTU = 0.0, CurTV = 0.0;


void d3dBegin( D3DPRIMITIVETYPE Type, IDirect3DDevice8* pDevice )
{
	//Initialize all our vars
	if( pDevice == NULL )
		return;
	pD3DDevice = pDevice;
	pD3DDevice->AddRef();	//Make sure our pointer doesn't suddenly go vanish
	
	PrimitiveType = Type;

	pVertList = new ListVertex[MIN_VERTEX_LIST_SIZE];
	NumVerts = 0;
	MaxArraySize = MIN_VERTEX_LIST_SIZE;
	
	CurDiffuse = 0xff000000;

	CurNX = 0.0; 
	CurNY = 0.0; 
	CurNZ = 0.0;
	
	CurTU = 0.0; 
	CurTV = 0.0;

	bRenderBegun = TRUE;
}


void d3dVertex3f( float x, float y, float z )
{
	//If we haven't begun, then there's no array to add to! :o
	if( !bRenderBegun )
		return;

	//If we're out of space in the array, add more space
	//Note: NumVerts starts from 1, but the array starts at 0. This means that
	//The next vertex will be added at pVertList[NumVerts]
	if( NumVerts == MaxArraySize )
	{
		//Expand the array by MIN_VERTEX_LIST_SIZE
		MaxArraySize += MIN_VERTEX_LIST_SIZE;
		ListVertex* pTemp = new ListVertex[MaxArraySize];

		memcpy( pTemp, pVertList, NumVerts * sizeof(ListVertex) );

		delete[] pVertList;
		pVertList = pTemp;
	}

	//Alias the current vertex to spare a bit of typing
	ListVertex* CurVert = &(pVertList[NumVerts]); //We haven't incremented NumVerts yet, so the last item is NumVerts, not NumVerts - 1
	
	CurVert->x = x;
	CurVert->y = y;
	CurVert->z = z;

	CurVert->nx = CurNX;
	CurVert->ny = CurNY;
	CurVert->nz = CurNZ;

	CurVert->Diffuse = CurDiffuse;
	
	CurVert->tu = CurTU;
	CurVert->tu = CurTV;

	NumVerts++;
}


//Takes a pointer to 3 floats. Useful if you like Carmack-style vertices(float Vertex[3])
void d3dVertex3fv( const float* pVector )
{
	d3dVertex3f( pVector[0], pVector[1], pVector[2] );
}


void d3dColor4f( float a, float r, float g, float b )
{
	CurDiffuse = D3DCOLOR_COLORVALUE( r, g, b, a );	//The param order for this macro is strange, but it comes out ARGB in the end, so it's all good
}


void d3dColor4fv( const float* pColor )
{
	d3dColor4f( pColor[0], pColor[1], pColor[2], pColor[3] );
}


void d3dColor3f( float r, float g, float b )
{
	CurDiffuse = D3DCOLOR_COLORVALUE( r, g, b, 0xff );	//Alpha value of 255
}


void d3dColor4ub( unsigned char a, unsigned char r, unsigned char g, unsigned char b )
{
	CurDiffuse = D3DCOLOR_ARGB( a, r, g, b );
}


void d3dColor3ub( unsigned char r, unsigned char g, unsigned char b )
{
	CurDiffuse = D3DCOLOR_XRGB( r, g, b );
}


void d3dNormal3f( float nx, float ny, float nz )
{
	CurNX = nx;
	CurNY = ny;
	CurNZ = nz;
}


void d3dTexCoord2f( float tu, float tv )
{
	CurTU = tu;
	CurTV = tv;
}


//Now, the most complex and possibly most important function
void d3dEnd()
{
	HRESULT r = 0;

	if( !bRenderBegun )
		return;

	int NumPrimitives = 0;
	//We need to calculate the number of faces, based on the primitive type requested
	switch( PrimitiveType )
	{
	case D3DPT_POINTLIST:
		NumPrimitives = NumVerts;
		break;
	case D3DPT_LINELIST:
		NumPrimitives = NumVerts / 2;
		break;
	case D3DPT_LINESTRIP:
		NumPrimitives = NumVerts - 1;
		break;
	case D3DPT_TRIANGLELIST:
		NumPrimitives = NumVerts / 3;
		break;
	case D3DPT_TRIANGLESTRIP:
		NumPrimitives = NumVerts - 2;
		break;
	case D3DPT_TRIANGLEFAN:
		NumPrimitives = NumVerts - 2;
	}


	//Create a vertex buffer and fill it
	IDirect3DVertexBuffer8* pVB = NULL;

	r = pD3DDevice->CreateVertexBuffer( sizeof(ListVertex) * NumVerts, D3DUSAGE_WRITEONLY, 
										FVF_LISTVERTEX, D3DPOOL_DEFAULT, &pVB );
	if( FAILED( r ) )
	{
		//Don't forget that there are things to do before bailing!
		pD3DDevice->Release();
		delete[] pVertList;
		bRenderBegun = FALSE;

		return;
	}

	void* pVertexData = NULL;

	r = pVB->Lock( 0, 0, (BYTE**)&pVertexData, 0 );
	if( FAILED( r ) )
	{
		pVB->Release();
		pD3DDevice->Release();
		delete[] pVertList;
		bRenderBegun = FALSE;

		return;
	}

	memcpy( pVertexData, pVertList, sizeof(ListVertex) * NumVerts );

	pVB->Unlock();

	pD3DDevice->SetStreamSource( 0, pVB, sizeof(ListVertex) );
	pD3DDevice->SetVertexShader( FVF_LISTVERTEX );
	r = pD3DDevice->DrawPrimitive( PrimitiveType, 0, NumPrimitives );

	//release stuff, delete memory, etc.
	pVB->Release();
	pD3DDevice->Release();
	delete[] pVertList;
	pVertList = NULL;
	bRenderBegun = FALSE;
}